library staterepository;

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';

// Should inherit from NotifiableState to receive Change Notifications from Repositories
// The only reason a create this class is because setState is protected
abstract class NotifiableState<T extends StatefulWidget> extends State<T> {
  void notifyMe(VoidCallback fn) {
    setState(fn);
  }
}

class RepositoryException implements Exception {
  String cause;
  RepositoryException(this.cause);
}

// Inherit from Repository to create your Data Repository
class Repository {
  static Map<String, List<Repository>> instances =
      Map<String, List<Repository>>();
  ObserverList<NotifiableState> _listeners = ObserverList<NotifiableState>();

  // Returns the repository of context
  // if no context id given then returns the first (default) Repository of type T in the list
  static T of<T extends Repository>([NotifiableState context]) {
    List<Repository> _classInstances =
        Repository.getClassIntances(T.toString());
    T defaultRepo;
    if (_classInstances.length > 0) {
      defaultRepo = _classInstances[0];
      if (context != null)
        for (Repository repository in _classInstances) {
          if (repository._listeners.contains(context)) return repository as T;
        }
    } else
      throw new RepositoryException(
          'There is no one repository of ${T.toString()} created.');
    return defaultRepo;
  }

  // sign for notification when default Repository of type T has change

  static void notifyMeChangesOf<T extends Repository>(NotifiableState context,
      [T instance]) {
    T defaultRepo = instance == null ? Repository.of<T>() : instance;
    if (defaultRepo != null)
      defaultRepo.addListener(context);
    else
      throw new RepositoryException(
          'There is no one repository of ${T.toString()} created.');
  }

//Add a listener
  void addListener(NotifiableState listener) {
    _listeners.add(listener);
  }

//remove a listener
  void removeListener(listener) {
    _listeners.remove(listener);
  }

// Should call notifyListeners(); in methods/properties
//void increment()
//{
//  count++;
//  notifyListeners();
// }
  void notifyListeners() {
    if (_listeners != null) {
      final List<NotifiableState> localListeners =
          List<NotifiableState>.from(_listeners);
      for (NotifiableState listener in localListeners) {
        listener.notifyMe(() {
          print(
              'Repository ${this.runtimeType.toString()} notifying ${listener.runtimeType.toString()}');
        });
      }
    }
  }

  static List<Repository> getClassIntances(String className) {
    List<Repository> _classInstances;
    //check if exists this class on map
    if (Repository.instances.containsKey(className))
      _classInstances = Repository.instances[className];
    else {
      _classInstances = List<Repository>();
      Repository.instances[className] = _classInstances;
    }
    return _classInstances;
  }

  Repository() {
    List<Repository> _classInstances =
        Repository.getClassIntances(this.runtimeType.toString());
    _classInstances.add(this);
  }

  void dispose() {
    List<Repository> _classInstances =
        Repository.getClassIntances(this.runtimeType.toString());
    _classInstances.remove(this);
  }
}
